/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.example.server.r4;

import ca.uhn.fhir.parser.DataFormatException;
import ca.uhn.fhir.rest.annotation.*;
import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.server.IResourceProvider;
import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.fhir.rest.server.exceptions.ResourceNotFoundException;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.InstantType;
import org.hl7.fhir.r4.model.Observation;

import java.util.*;

public class ObservationResourceProvider implements IResourceProvider {
	/**
	 * This map has a resource ID as a key, and each key maps to a Deque list containing all versions of the resource with that ID.
	 */
	private Map<Long, Deque<Observation>> myIdToObservationVersions = new HashMap<>();
	/**
	 * This is used to generate new IDs
	 */
	private long myNextId = 1L;

	/**
	 * Constructor, which pre-populates the provider with one resource instance.
	 */
	public ObservationResourceProvider() {
		long resourceId = myNextId++;

		Observation obs = new Observation();
		obs.setId(Long.toString(resourceId));
		obs.addIdentifier();
		obs.getIdentifier().get(0).setSystem("urn:hapitest:obs");
		obs.getIdentifier().get(0).setValue("00002");
		obs.setStatus(Observation.ObservationStatus.FINAL);

		LinkedList<Observation> list = new LinkedList<>();
		list.add(obs);

		myIdToObservationVersions.put(resourceId, list);
	}

	/**
	 * Stores a new version of the Observation in memory so that it can be retrieved later.
	 *
	 * @param theObservation The Observation resource to store
	 * @param theId          The ID of the Observation to retrieve
	 */
	private void addNewVersion(Observation theObservation, Long theId) {
		if (!myIdToObservationVersions.containsKey(theId)) {
			myIdToObservationVersions.put(theId, new LinkedList<>());
		}

		theObservation.getMeta().setLastUpdatedElement(InstantType.withCurrentTime());

		Deque<Observation> existingVersions = myIdToObservationVersions.get(theId);

		// We just use the current number of versions as the next version number
		String newVersion = Integer.toString(existingVersions.size());

		// Create an ID with the new version and assign it back to the resource
		IdType newId = new IdType("Observation", Long.toString(theId), newVersion);
		theObservation.setId(newId);

		existingVersions.add(theObservation);
	}

	/**
	 * The "@Create" annotation indicates that this method implements "create=type", which adds a
	 * new instance of a resource to the server.
	 */
	@Create()
	public MethodOutcome createObservation(@ResourceParam Observation theObservation) {

		// Here we are just generating IDs sequentially
		long id = myNextId++;

		addNewVersion(theObservation, id);

		// Let the caller know the ID of the newly created resource
		return new MethodOutcome(new IdType(id));
	}

	@Search
	public List<Observation> findObservationsUsingArbitraryCtriteria() {
		LinkedList<Observation> retVal = new LinkedList<>();

		for (Deque<Observation> nextObservationList : myIdToObservationVersions.values()) {
			Observation nextObservation = nextObservationList.getLast();
			retVal.add(nextObservation);
		}

		return retVal;
	}

	/**
	 * The getResourceType method comes from IResourceProvider, and must be overridden to indicate what type of resource this provider supplies.
	 */
	@Override
	public Class<Observation> getResourceType() {
		return Observation.class;
	}

	/**
	 * This is the "read" operation. The "@Read" annotation indicates that this method supports the read and/or vread operation.
	 * <p>
	 * Read operations take a single parameter annotated with the {@link IdParam} paramater, and should return a single resource instance.
	 * </p>
	 *
	 * @param theId The read operation takes one parameter, which must be of type IdDt and must be annotated with the "@Read.IdParam" annotation.
	 * @return Returns a resource matching this identifier, or null if none exists.
	 */
	@Read(version = true)
	public Observation readObservation(@IdParam IdType theId) {
		Deque<Observation> retVal;
		try {
			retVal = myIdToObservationVersions.get(theId.getIdPartAsLong());
		} catch (NumberFormatException e) {
			/*
			 * If we can't parse the ID as a long, it's not valid so this is an unknown resource
			 */
			throw new ResourceNotFoundException(theId);
		}

		if (!theId.hasVersionIdPart()) {
			return retVal.getLast();
		} else {
			for (Observation nextVersion : retVal) {
				String nextVersionId = nextVersion.getIdElement().getVersionIdPart();
				if (theId.getVersionIdPart().equals(nextVersionId)) {
					return nextVersion;
				}
			}
			// No matching version
			throw new ResourceNotFoundException("Unknown version: " + theId.getValue());
		}
	}

	/**
	 * The "@Update" annotation indicates that this method supports replacing an existing
	 * resource (by ID) with a new instance of that resource.
	 *
	 * @param theId          This is the ID of the Observation to update
	 * @param theObservation This is the actual resource to save
	 * @return This method returns a "MethodOutcome"
	 */
	@Update()
	public MethodOutcome updateObservation(@IdParam IdType theId, @ResourceParam Observation theObservation) {

		Long id;
		try {
			id = theId.getIdPartAsLong();
		} catch (DataFormatException e) {
			throw new InvalidRequestException("Invalid ID " + theId.getValue() + " - Must be numeric");
		}

		/*
		 * Throw an exception (HTTP 404) if the ID is not known
		 */
		if (!myIdToObservationVersions.containsKey(id)) {
			throw new ResourceNotFoundException(theId);
		}

		addNewVersion(theObservation, id);

		return new MethodOutcome();
	}
}
