/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.example.server.r4;

import ca.uhn.fhir.batch2.jobs.config.Batch2JobsConfig;
import ca.uhn.fhir.context.FhirContext;
import ca.uhn.fhir.jpa.api.config.JpaStorageSettings;
import ca.uhn.fhir.jpa.api.config.ThreadPoolFactoryConfig;
import ca.uhn.fhir.jpa.batch2.JpaBatch2Config;
import ca.uhn.fhir.jpa.binary.api.IBinaryStorageSvc;
import ca.uhn.fhir.jpa.binstore.MemoryBinaryStorageSvcImpl;
import ca.uhn.fhir.jpa.config.HapiJpaConfig;
import ca.uhn.fhir.jpa.config.r4.JpaR4Config;
import ca.uhn.fhir.jpa.config.util.HapiEntityManagerFactoryUtil;
import ca.uhn.fhir.jpa.model.config.PartitionSettings;
import ca.uhn.fhir.jpa.model.dialect.HapiFhirH2Dialect;
import ca.uhn.fhir.jpa.subscription.channel.config.SubscriptionChannelConfig;
import ca.uhn.fhir.jpa.subscription.match.config.SubscriptionProcessorConfig;
import ca.uhn.fhir.jpa.subscription.submit.config.SubscriptionSubmitterConfig;
import ca.uhn.fhir.jpa.util.CircularQueueCaptureQueriesListener;
import ca.uhn.fhir.jpa.util.CurrentThreadCaptureQueriesListener;
import ca.uhn.fhir.rest.server.FifoMemoryPagingProvider;
import ca.uhn.fhir.rest.server.IPagingProvider;
import com.example.server.common.TestJpaConfig;
import jakarta.persistence.EntityManagerFactory;
import net.ttddyy.dsproxy.listener.SingleQueryCountHolder;
import net.ttddyy.dsproxy.listener.logging.SLF4JLogLevel;
import net.ttddyy.dsproxy.support.ProxyDataSourceBuilder;
import org.apache.commons.dbcp2.BasicDataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.Primary;
import org.springframework.orm.jpa.JpaTransactionManager;
import org.springframework.orm.jpa.LocalContainerEntityManagerFactoryBean;

import java.util.Properties;
import java.util.concurrent.TimeUnit;
import javax.sql.DataSource;

/**
 * Optional configuration to set up DaoRegistry and Dao access for Tester class, can be removed, if not required.
 * It uses an internal in-memory H2 database to keep FHIR data temporarily,
 * to demonstrate how to configure access to FHIR database.
 * Ultimately, it could be configured to use the same database that is configured
 * in a Smile CDR Persistence module, in a development or test environment, for example.
 */
@Configuration
@Import({
	JpaR4Config.class,
	HapiJpaConfig.class,
	JpaBatch2Config.class,
	Batch2JobsConfig.class,
	SubscriptionSubmitterConfig.class,
	SubscriptionProcessorConfig.class,
	SubscriptionChannelConfig.class,
	ThreadPoolFactoryConfig.class
})
public class TestJpaR4Config extends TestJpaConfig {

	private static final Logger ourLog = LoggerFactory.getLogger(TestJpaR4Config.class);

	private static final int ourMaxThreads = 20;

	@Bean
	public CircularQueueCaptureQueriesListener captureQueriesListener() {
		return new CircularQueueCaptureQueriesListener();
	}

	@Bean
	public DataSource dataSource() {
		BasicDataSource retVal = new BasicDataSource();

		retVal.setDriver(new org.h2.Driver());
		retVal.setUrl("jdbc:h2:mem:testdb_r4");

		retVal.setUsername("h2InMemory");
		retVal.setPassword("SmileCDR");

		retVal.setMaxWaitMillis(30000);
		retVal.setMaxTotal(ourMaxThreads);

		SLF4JLogLevel logLevel = SLF4JLogLevel.INFO;
		return ProxyDataSourceBuilder.create(retVal)
				//			.logQueryBySlf4j(level)
				.logSlowQueryBySlf4j(10, TimeUnit.SECONDS, logLevel)
				.afterQuery(captureQueriesListener())
				.afterQuery(new CurrentThreadCaptureQueriesListener())
				.countQuery(singleQueryCountHolder())
				.afterMethod(captureQueriesListener())
				.build();
	}

	@Bean
	public SingleQueryCountHolder singleQueryCountHolder() {
		return new SingleQueryCountHolder();
	}

	@Bean
	public LocalContainerEntityManagerFactoryBean entityManagerFactory(
			ConfigurableListableBeanFactory theConfigurableListableBeanFactory,
			FhirContext theFhirContext,
			JpaStorageSettings theStorageSettings) {

		LocalContainerEntityManagerFactoryBean retVal = HapiEntityManagerFactoryUtil.newEntityManagerFactory(
				theConfigurableListableBeanFactory, theFhirContext, theStorageSettings);
		retVal.setPersistenceUnitName("PU_HapiFhirJpaR4");
		retVal.setDataSource(dataSource());
		retVal.setJpaProperties(jpaProperties());
		return retVal;
	}

	private Properties jpaProperties() {
		Properties extraProperties = new Properties();
		extraProperties.put("hibernate.search.enabled", "false");
		extraProperties.put("hibernate.format_sql", "false");
		extraProperties.put("hibernate.show_sql", "false");
		extraProperties.put("hibernate.hbm2ddl.auto", "update");
		extraProperties.put("hibernate.dialect", HapiFhirH2Dialect.class.getName());

		ourLog.info("jpaProperties: {}", extraProperties);

		return extraProperties;
	}

	@Bean
	public JpaStorageSettings storageSettings() {
		JpaStorageSettings storageSettings = new JpaStorageSettings();

		// set it to generate UUID as Fhir id, server side
		storageSettings.setResourceServerIdStrategy(JpaStorageSettings.IdStrategyEnum.UUID);

		// allow :missing searches
		storageSettings.setIndexMissingFields(JpaStorageSettings.IndexEnabledEnum.ENABLED);

		return storageSettings;
	}

	@Bean
	public PartitionSettings partitionSettings() {
		return new PartitionSettings();
	}

	@Bean
	@Primary
	public JpaTransactionManager transactionManager(EntityManagerFactory entityManagerFactory) {
		JpaTransactionManager retVal = new JpaTransactionManager();
		retVal.setEntityManagerFactory(entityManagerFactory);
		return retVal;
	}

	@Bean
	@Lazy
	public IBinaryStorageSvc binaryStorage() {
		return new MemoryBinaryStorageSvcImpl();
	}

	@Bean
	@Primary
	IPagingProvider pagingProvider() {
		FifoMemoryPagingProvider retVal = new FifoMemoryPagingProvider(10);
		retVal.setDefaultPageSize(10);
		retVal.setMaximumPageSize(100);
		return retVal;
	}
}
